local NAME, addon = ...
addon.NAME = NAME
XLootGroup = addon

local L = addon.L
local me = UnitName('player')

local opt, roll_anchor, watch_anchor
-- Upvalue for use in testing (at end)
local GetLootRollItemInfo, GetLootRollItemLink, RollOnLoot =
	GetLootRollItemInfo, GetLootRollItemLink, RollOnLoot

local defaults = {
	text_outline = true,

	roll_button_size = 28,
	roll_width = 325,

	roll_anchor = {
		direction = 'up',
		visible = true,
		scale = 1.0,
		x = UIParent:GetWidth() * .85,
		y = UIParent:GetHeight() * .6
	},
	watch_enable = true,
	watch_anchor = {
		direction = 'up',
		visible = true,
		scale = 1.0,
		x = UIParent:GetWidth() * .85,
		y = UIParent:GetHeight() * .5
	}
}
addon.defaults = defaults
local print = print

--// SKINNING
local Skinner = {}
LibStub('X-Skin'):SetupSkins(Skinner, {
	row = {},
	item = {},
	anchor = {}
})

local skin_default = {
	name = ('|c2244dd22%s|r'):format('Smooth'),
	texture = [[Interface\AddOns\XLootGroup\textures\border_smooth]],
	bar_texture = [[Interface\AddOns\XLootGroup\textures\bar]],
	color_mod = .85,
	size = 14,
	padding = 1,
}

local skin
local skins_base = {
	anchor = { r = .6, g = .6, b = .6, a = .8 },
	row = { gradient = false },
	item = { backdrop = false }
}

local function CompileSkins()
	local skins = Skinner._skins
	if XLootFrame and XLootFrame.skin then
		skin = XLootFrame.skin
	else
		skin = skin_default
	end
	for which,subskin in pairs(skins) do
		wipe(subskin)
		for k,v in pairs(skin_default) do
			subskin[k] = skin[k]
		end
		for k,v in pairs(skins_base[which]) do
			subskin[k] = v
		end
	end
end

-- Since # doesn't work for non-contiguous tables
local function count(table)
	local i = 0
	for k,v in pairs(table) do
		i = i + 1
	end
	return i
end

-- Tack role on to player name and return class colors
local white = { r = 1, g = 1, b = 1 }
local function FancyPlayerName(name)
	local c = RAID_CLASS_COLORS[select(2, UnitClass(name))] or white
	local role = UnitGroupRolesAssigned(name)
	if role ~= 'NONE' then
		name = ('%s (%s)'):format(name, _G[role])
	end
	return name, c.r, c.g, c.b
end

--// FRAME CREATION
local ReskinBars, mouse_focus, CreateRollFrame, CreateWatchFrame
do
	local bars = {}
	function ReskinBars()
		local pad = skin.padding or skin_default.padding
		local tex = skin.bar_texture or skin_default.bar_texture
		local p, n = pad + 3, -pad - 3
		for _,bar in pairs(bars) do
			bar:ClearAllPoints()
			bar:SetPoint('TOPRIGHT', n, n)
			bar:SetPoint('BOTTOMRIGHT', n, p)
			bar:SetPoint('LEFT', bar.parent.icon_frame, 'RIGHT', -pad, 0)
			bar:SetStatusBarTexture(tex)
			local link = bar.parent.link
			if link then
				local r, g, b = GetItemQualityColor(select(3, GetItemInfo(link)))
				bar.parent.overlay:SetBorderColor(r, g, b)
				bar.parent.icon_frame:SetBorderColor(r, g, b)
			end
		end
	end

	-- General item events
	local function OnEnter(self)
		mouse_focus = self
		GameTooltip:SetOwner(self.icon_frame, 'ANCHOR_TOPLEFT', 28, 0)
		GameTooltip:SetHyperlink(self.link)
		if IsShiftKeyDown() then
			GameTooltip_ShowCompareItem()
		end
		if IsModifiedClick('DRESSUP') then
			ShowInspectCursor()
		else
			ResetCursor()
		end
	end

	local function OnLeave(self)
		mouse_focus = nil
		GameTooltip:Hide()
	end

	local function OnClick(self, button)
		if IsControlKeyDown() then
			DressUpItemLink(self.link)
		elseif IsShiftKeyDown() then
			ChatEdit_InsertLink(self.link)
		end
	end

	-- Update all status bars
	local max = math.max
	local function BarUpdate(self)
		local now, length = max(self.expires - GetTime(), -1), self.length
		self.spark:SetPoint('CENTER', self, 'LEFT', (now / length) * self:GetWidth(), 0)
		self:SetValue(now)
	end

	-- Fill item information on rows
	local function UpdateRow(self, id, link, icon, name, bop, r, g, b, start, length)
		local bar = self.bar
		bar.length = length
		bar.expires = GetTime() + start

		self.link = link
		self.rollid = id

		self.text_bind:SetText(bop and '|cffff4422BoP' or '')
		self.text_loot:SetText(name)

		self.text_loot:SetVertexColor(r, g, b)
		self.overlay:SetBorderColor(r, g, b)
		self.icon_frame:SetBorderColor(r, g, b)
		bar:SetStatusBarColor(r, g, b, .7)
		self.icon:SetTexture(icon)

		bar:SetMinMaxValues(0, length)
		bar:SetValue(start)
	end

	-- Base rows
	local function CreateRow()
		local frame = CreateFrame('Button', nil, UIParent)
		frame:SetFrameLevel(roll_anchor:GetFrameLevel())
		frame:SetHeight(24)
		frame:SetWidth(opt.roll_width)
		frame:SetScript('OnEnter', OnEnter)
		frame:SetScript('OnLeave', OnLeave)
		frame:SetScript('OnClick', OnClick)
		frame.OnEnter = OnEnter
		frame.Start = Start
		
		local overlay = CreateFrame('frame', nil, frame)
		overlay:SetFrameLevel(frame:GetFrameLevel())
		overlay:SetAllPoints()
		Skinner:Skin(overlay, 'row')

		local icon_frame = CreateFrame('Frame', nil, frame)
		icon_frame:SetPoint('LEFT', 0, 0)
		icon_frame:SetWidth(28)
		icon_frame:SetHeight(28)
		Skinner:Skin(icon_frame, 'item')

		local icon = icon_frame:CreateTexture(nil, 'BACKGROUND')
		icon:SetPoint('TOPLEFT', 3, -3)
		icon:SetPoint('BOTTOMRIGHT', -3, 3)
		icon:SetTexCoord(.07,.93,.07,.93)
		
		local bar = CreateFrame('StatusBar', nil, frame)
		bar:SetFrameLevel(frame:GetFrameLevel())
		local pad = skin.padding or skin_default.padding
		bar:SetPoint('TOPRIGHT', -pad - 3, -pad - 3)
		bar:SetPoint('BOTTOMRIGHT', -pad - 3, pad + 3)
		bar:SetPoint('LEFT', icon_frame, 'RIGHT', -pad, 0)
		bar:SetStatusBarTexture(skin.bar_texture or skin_default.bar_texture)
		bar:SetScript('OnUpdate', BarUpdate)
		bar.parent = frame
		table.insert(bars, bar)
		
		local spark = bar:CreateTexture(nil, 'OVERLAY')
		spark:SetWidth(14)
		spark:SetHeight(38)
		spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
		spark:SetBlendMode('ADD')
		bar.spark = spark

		local bind = icon_frame:CreateFontString(nil, 'OVERLAY')
		bind:SetPoint('BOTTOM', 0, 1)
		bind:SetFont([[Fonts\FRIZQT__.TTF]], 8, 'THICKOUTLINE')
		
		local loot = frame:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
		loot:SetHeight(16)
		loot:SetJustifyH('LEFT')
		if opt.text_outline then
			local f, s = loot:GetFont()
			loot:SetFont(f, s, 'OUTLINE')
		end
		loot:SetPoint('RIGHT', frame, 'RIGHT', -5, 0)

		frame.text_bind = bind
		frame.text_loot = loot
		frame.overlay = overlay
		frame.bar = bar
		frame.icon = icon
		frame.icon_frame = icon_frame
		frame.Update = UpdateRow

		return frame
	end

	-- N/G/D/P buttons
	local function OnButtonClick(self)
		RollOnLoot(self.parent.rollid, self.type)
	end
	
	local function Toggle(self, status)
		if status then
			self:Enable()
			self:SetAlpha(1)
		else
			self:Disable()
			self:SetAlpha(.6)
		end
		SetDesaturation(self:GetNormalTexture(), not status)
	end

	local type_index = { 'need', 'greed', 'disenchant', [0] = 'pass' }
	local function OnButtonEnter(self)
		if self:IsEnabled() == 0 then return end
		mouse_focus = self
		GameTooltip:SetOwner(self, 'ANCHOR_TOPLEFT')
		GameTooltip:SetText(self.label)
		local rtype = type_index[self.type]
		for name,roll in pairs(self.parent.rolls) do
			if roll == rtype then
				GameTooltip:AddLine(FancyPlayerName(name))
			end
		end
		GameTooltip:Show()
	end

	local function OnButtonLeave(self)
		GameTooltip:Hide()
	end

	-- TODO: Automatically update tooltip?
	local path = [[Interface\Buttons\UI-GroupLoot-%s-%s]]
	local function CreateRollButton(parent, roll, label, tex, to, x, y)
		local b = CreateFrame('Button', nil, parent)
		b:SetPoint('LEFT', to, 'RIGHT', x, y)
		b:SetWidth(opt.roll_button_size)
		b:SetHeight(opt.roll_button_size)
		b:SetNormalTexture(path:format(tex, 'Up'))
		if tex ~= 'Pass' then
			b:SetHighlightTexture(path:format(tex, 'Highlight'))
			b:SetPushedTexture(path:format(tex, 'Down'))
		else
			b:SetHighlightTexture(path:format(tex, 'Up'))
			b:GetNormalTexture():SetVertexColor(0.8, 0.7, 0.7)
			b:GetHighlightTexture():SetAlpha(0.5)
		end
		b.parent = parent

		local text = b:CreateFontString(nil, 'OVERLAY')
		text:SetFont([[Fonts\FRIZQT__.TTF]], 12, 'THICKOUTLINE')
		text:SetPoint("CENTER", -x + 1, -y)
		b.text = text

		b:SetScript('OnEnter', OnButtonEnter)
		b:SetScript('OnLeave', OnButtonLeave)
		b:SetScript('OnClick', OnButtonClick)
		b:SetMotionScriptsWhileDisabled(true)
		b.OnEnter = OnButtonEnter
		b.Toggle = Toggle
		b.type = roll
		b.label = label

		return b
	end

	local function RollPop(self)
		wipe(self.rolls)
		self.rollid = nil
		self.link = nil
	end

	-- Group roll frame
	function CreateRollFrame()
		local frame = CreateRow()
		frame.Popped = RollPop

		local n = CreateRollButton(frame, 1, NEED, 'Dice', frame.icon_frame, 3, -1)
		local g = CreateRollButton(frame, 2, GREED, 'Coin', n, 0, -2)
		local d = CreateRollButton(frame, 3, ROLL_DISENCHANT, 'DE', g, 0, 2)
		local p = CreateRollButton(frame, 0, PASS, 'Pass', d, 0, 2)

		frame.text_loot:SetPoint('LEFT', p, 'RIGHT', 3, 0)

		frame.need, frame.greed, frame.disenchant, frame.pass = n, g, d, p
		frame.rolls = {}

		return frame
	end

	-- Watch frames
	local function lines(players)
		local me = UnitName('player')
		local sf = string.format
		for player, roll in pairs(players) do
			local text, r, g, b, color = FancyPlayerName(player)
			if roll ~= true then
				if mouse_focus.winner == player then
					color = '44ff22'
				elseif player == me then
					color = 'ff2244'
				else
					color = 'CCCCCC'
				end
				GameTooltip:AddLine(sf('   |cff%s%s|r  %s', color, roll, player), r, g, b)
			else
				GameTooltip:AddLine('   '..player, r, g, b)
			end
		end
	end

	local function WatchEnter(self)
		OnEnter(self)
		if next(self.rolls) then
			GameTooltip:AddLine('.', 0, 0, 0)
		end
		local rolls_type = self.rolls_type
		if count(rolls_type.need) > 0 then
			GameTooltip:AddLine(NEED, .2, 1, .1)
			lines(rolls_type.need)
		end
		if count(rolls_type.greed) > 0 then
			GameTooltip:AddLine(GREED, .1, .2, 1)
			lines(rolls_type.greed)
		end
		if count(rolls_type.disenchant) > 0 then
			GameTooltip:AddLine(ROLL_DISENCHANT, 1, .2, 1)
			lines(rolls_type.disenchant)
		end
		if count(rolls_type.pass) > 0 then
			GameTooltip:AddLine(PASS, .7, .7, .7)
			lines(rolls_type.pass)
		end
		GameTooltip:Show()
	end

	local function WatchPop(self)
		RollPop(self)
		for _, t in pairs(self.rolls_type) do
			wipe(t)
		end
		self.winner = nil
		self.text:SetText()
	end

	local function UpdateText(self)
		self.text_loot:SetWidth(self.max_text - self.text:GetStringWidth() - 8)
	end

	function CreateWatchFrame()
		local frame = CreateRow()
		frame:SetScript('OnEnter', WatchEnter)
		frame:SetWidth(235)
		frame.OnEnter = WatchEnter
		frame.Popped = WatchPop
		frame.UpdateText = UpdateText

		local text = frame:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
		text:SetHeight(16)
		text:SetJustifyH('LEFT')
		if opt.text_outline then
			local f, s = text:GetFont()
			text:SetFont(f, s, 'OUTLINE')
		end
		text:SetPoint('RIGHT', frame, 'RIGHT', -5, 0)
		text:SetTextColor(.7, .7, .7)

		frame.text_loot:ClearAllPoints()
		frame.text_loot:SetPoint('LEFT', frame.icon_frame, 'RIGHT', 3, 0)
		--frame.text_loot:SetPoint('RIGHT', text, 'RIGHT', -15, 0)
		frame.max_text = frame:GetWidth() - frame.icon_frame:GetWidth()

		frame.text = text
		frame.rolls = {}
		frame.rolls_type = {
			need = {},
			greed = {},
			disenchant = {},
			pass = {}
		}

		return frame
	end
end

--// EVENT HANDLING
local rolls, watches, over = { }, { }, { }
-- Retrieve frame by item link
local function find_frame(frames, item, who, rolled)
	for _,f in pairs(frames) do
		if f.link == item and ((rolled and f.rolls[who]) or (not rolled and not f.rolls[who])) then
			return f
		end
	end
end

local types = {
	need = NEED,
	greed = GREED,
	disenchant = ROLL_DISENCHANT,
	pass = PASS
}
local sf = string.format
local function rollevent(event, chat_event, ...)
	local who, item, rtype, roll
	local name = GetItemInfo((...))
	--print(event, chat_event, name, select(2, ...))
	-- Incriment counter on roll frame
	if event == 'selected' then
		item, who, rtype = ...
		-- Update roll frame
		local f = find_frame(rolls, item, who)
		if f then
			f.rolls[who] = rtype
			-- Button text
			local counter = f[rtype].text
			counter:SetText(counter:GetText() and counter:GetText() + 1 or 1)
			-- Update tooltip
			if f[rtype] == mouse_focus then
				mouse_focus:OnEnter()
			end
		end
	end

	if opt.watch_enable then
		-- Save choice
		local watch
		if event == 'selected' then
			item, who, rtype = ...
			watch = find_frame(watches, item, who)
			-- Tracking
			local rolls, rolls_type = watch.rolls, watch.rolls_type
			watch.rolls[who] = rtype
			rolls_type[rtype][who] = true
			-- Watch text
			local num
			local n, g, d, p = count(rolls_type.need), count(rolls_type.greed), count(rolls_type.disenchant), count(rolls_type.pass)
			if n > 0 then
				num = n
				rtype = 'need'
			elseif g > 0 or d > 0 then
				num = g + d
				rtype = 'greed'
			elseif p > 0 then
				num = p
				rtype = 'pass'
			end
			local r, g, b = .7, .7, .7
			if rolls[me] == rtype then
				r, g, b = .2, 1, .1
			elseif rolls[me] and rolls[me] ~= 'pass' then
				r, g, b = 1, .2, .1
			end
			watch.text:SetText(sf('%s: %s |cff777777...', types[rtype], num))
			watch.text:SetTextColor(r, g, b)

		-- Count rolls 
		elseif event == 'rolled' then
			local item, who, rtype, roll = ...
			watch = find_frame(watches, item, who, true)
			watch.rolls_type[rtype][who] = roll
			-- Refresh tooltip
			if mouse_focus == watch then
				watch:OnEnter()
			end

		-- Conclude watches
		elseif event == 'won' then
			local item, winner = ...
			watch = find_frame(watches, item, winner, true)
			local win_type = watch.rolls[winner]
			local bracket = watch.rolls_type[win_type]
			local win_roll, my_roll = bracket[winner], bracket[me]
			local player, r, g, b = FancyPlayerName(winner)

			local color, expire = '777777', 10
			if winner == me then
				color = '44ff22'
				expire = 20
			elseif watch.rolls[me] ~= 'pass' then
				color = 'ff2244'
			end

			text = winner
			-- if winner ~= me and my_roll then
			-- 	text = sf('%s|cff%s %s: %s/%s', player, color, types[win_type], win_roll, my_roll)
			-- 	expire = 20
			-- else
			-- 	text = sf('%s|cff%s %s: %s', player, color, types[win_type], win_roll)
			-- end

			watch.text:SetText(text)
			watch.text:SetTextColor(r, g, b)
			watch_anchor:Expire(watch, expire)
			watch.winner = winner
			watch.bar.expires = GetTime()

		-- Pass conclusion
		elseif event == 'allpass' then
			watch = find_frame(watches, ...)
			watch.text:SetText(sf('%s: %s', PASS, ALL))
			watch.text:SetTextColor(.7, .7, .7)
			watch_anchor:Expire(watch, 10)
			watch.bar.expires = GetTime()
		end
		-- Refresh tooltip
		if watch then
			if mouse_focus == watch then
				watch:OnEnter()
			end
			watch:UpdateText()
		end
	end
end

local function start(id, length, ongoing)
	if id and over[id] then return end -- Failsafe?
	
	local icon, name, count, quality, bop, need, greed, de = GetLootRollItemInfo(id)
	local link = GetLootRollItemLink(id)
	local r, g, b = GetItemQualityColor(quality)

	local start = length
	if ongoing then
		if quality == 2 then
			length = 60000
		else
			length = 180000
		end
	end
	length, start = length/1000, start/1000

	local roll = roll_anchor:Push()
	roll:Update(id, link, icon, name, bop, r, g, b, start, length)
	rolls[id] = roll

	roll.need:Toggle(need)
	roll.greed:Toggle(greed)
	roll.disenchant:Toggle(de)

	roll.need.text:SetText()
	roll.greed.text:SetText()
	roll.pass.text:SetText()
	roll.disenchant.text:SetText()

	if opt.watch_enable then
		local watch = watch_anchor:Push()
		watch:Update(id, link, icon, name, bop, r, g, b, start, length)
		watches[id] = watch
		watch:UpdateText()
	end
	return roll
end

local function cancel(id)
	over[id] = true

	local f = rolls[id]
	if f then
		roll_anchor:Pop(f)
		rolls[id] = nil
	end
end

-- To avoid a OnUpdate for every mouseoverable frame, use modifier events
local function modifier(self, modifier, state)
	if mouse_focus and MouseIsOver(mouse_focus) and mouse_focus.OnEnter then
		mouse_focus:OnEnter()
	end
end

--// ADDON SETUP
LibStub('X-Profile'):Setup(addon, 'XLootGroup_Options')
function addon:OnProfileChanged(options)
	opt = options
	addon.opt = opt
end

-- Update skins when XLoot skin changes
function addon:SkinUpdate()
	CompileSkins()
	Skinner:Reskin()
	ReskinBars()
end

-- Move anchors when scale changes
do
	local function reposition(anchor)
		local mod = anchor:GetScale() / anchor.data.scale
		anchor:Scale(anchor.data.scale)
		anchor.data.x = anchor.data.x * mod
		anchor.data.y = anchor.data.y * mod
		anchor:Position()
		anchor:Restack()
	end

	function addon:ConfigSave()
		reposition(roll_anchor)
		reposition(watch_anchor)
	end
end

-- Remove watch frames after letting result linger
local timer = 0
local function OnWatchUpdate(self, elapsed)
	if timer < 2 then
		timer = timer + elapsed
	else
		timer = 0
		local time = GetTime()
		for i, watch in ipairs(watch_anchor.expiring) do
			if time > watch.expires then
				watch_anchor:Pop(watch)
				table.remove(watch_anchor.expiring, i)
			end
		end
	end
end

function addon:OnLoad()
	XStack = LibStub('X-Stack')

	-- Create Roll anchor
	roll_anchor = XStack:CreateStaticStack(CreateRollFrame, addon.L.roll_anchor, opt.roll_anchor)
	roll_anchor:SetFrameLevel(7)
	roll_anchor:Scale(opt.roll_anchor.scale)
	addon.roll_anchor = roll_anchor

	-- Create Watch anchor
	watch_anchor = XStack:CreateStaticStack(CreateWatchFrame, addon.L.watch_anchor, opt.watch_anchor)
	watch_anchor:SetFrameLevel(5)
	watch_anchor:Scale(opt.watch_anchor.scale)
	addon.watch_anchor = watch_anchor
	local pip_frame = CreateFrame('Frame')
	pip_frame:SetScript('OnUpdate', OnWatchUpdate)
	-- Watch linger/expire
	watch_anchor.expiring = {}
	function watch_anchor:Expire(watch, expires)
		table.insert(self.expiring, watch)
		watch.expires = GetTime() + expires
	end
	
	roll_anchor:RegisterEvent('START_LOOT_ROLL')
	roll_anchor:RegisterEvent('CANCEL_LOOT_ROLL')
	roll_anchor:RegisterEvent('MODIFIER_STATE_CHANGED')
	roll_anchor:SetScript('OnEvent', function(_, e, ...)
		if e == 'MODIFIER_STATE_CHANGED' then
			modifier(...)
		elseif e == 'START_LOOT_ROLL' then
			start(...)
		elseif e == 'CANCEL_LOOT_ROLL' then
			cancel(...)
		end
	end)

	UIParent:UnregisterEvent("START_LOOT_ROLL")
	UIParent:UnregisterEvent("CANCEL_LOOT_ROLL")

	LibStub('LootEvents'):RegisterGroupCallback(rollevent)
	LibStub('X-Config'):Setup(addon, 'XLootGroupOptionPanel')
	
	--// XLOOT REGISTRATION
	if XLootFrame and XLootFrame.RegisterPlugin then
		XLootFrame:RegisterPlugin(addon)
	end

	--// SKINNING
	CompileSkins()
	--Skinner:Skin(roll_anchor, 'anchor')

	--// ONGOING ROLLS
	if (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) and GetLootMethod() == 'group' then
		for i=1,200 do
			local time = GetLootRollTimeLeft(i)
			if time > 0 then
				start(i, time, true)
			end
		end
	end
end

--// SLASH COMMAND
local function option_handler(msg)
	if msg == 'reset' then
		addon:ResetProfile()
		addon.roll_anchor:Position()
	else
		if roll_anchor:IsShown() then
			roll_anchor:Hide()
			watch_anchor:Hide()
		else
			roll_anchor:Show()
			watch_anchor:Show()
		end
	end
end
SLASH_XLOOTGROUP1 = '/xlg'
SLASH_XLOOTGROUP2 = '/xlootgroup'
SlashCmdList['XLOOTGROUP'] = option_handler

